/*
Copyright 2005 Ethan McCallum (ethanqm -at- penguinmail.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

or in the file

	LICENSE

which is distributed with this package.

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
Limitations under the License.
*/

#include<support_classes.h>
#include<AppException.h>

#include<string>
#include<iostream>

// - - - - - - - - - - - - - - - - - - - -

extern "C" {
	#include<rpm/rpmlib.h>
	#include<rpm/rpmio.h>
	#include<rpm/rpmts.h>
	#include<rpm/header.h>
}

#include<cstdio>
#include<sstream>

RPM::RPM( const std::string& inFile )
	throw( AppException )
	:
	file_( inFile ) ,
	name_() ,
	arch_() ,
	version_() ,
	release_() ,
	fullVersion_() ,
	isSourcePackage_( false )
{

	/*
	QQQ: way to get just the headers?
	it takes ~12min to skim a few thousand RPMs.
	That's just silly.
	*/

	FD_t fileDescriptor = Fopen( file_.c_str() , "r.ufdio" ) ;

	Header hdr ;

	rpmts ts = rpmtsCreate() ;

	/*
	QQQ -- must forward past other info to get to header...
	don't see this much w/ non-distro RPMs, so the error
	appears intermittent, if not completely random
	*/

	// QQQ -- check for error (0 return val), throw exception
	// (get exception message from an immediate call to
	// "rpmlogMessage()"... "immediate" because this fn returns
	// the *last* RPM error.  Not quite thread-safe...

	const int retval = rpmReadPackageFile(
		ts,
		fileDescriptor ,
		file_.c_str() ,
		&hdr 
	) ;

	if( RPMRC_FAIL == retval ){
		std::cerr << "error loading " << file_ << ": " << rpmlogMessage() << std::endl ;
		throw( AppException( rpmlogMessage() ) ) ;
	} // switch( retval )


	// QQQ -- check for error (0 return val), throw exception
	void* headerEntry = NULL ;
	char* tmpString = NULL ;

	// QQQ - rewrite w/ headerIsEntry() to replace bulky tmp logic
	// (put needed tags into a collection, iterate through them
	//  all with headerIsEntry() first, then collect value w/
	//  headerGetEntry())
	headerGetEntry( hdr , RPMTAG_NAME , NULL , &headerEntry , NULL ) ;
	tmpString = static_cast< char* >( headerEntry ) ;

	if( NULL != tmpString ){
		name_ = tmpString ;
	}


	headerGetEntry( hdr , RPMTAG_VERSION , NULL , &headerEntry , NULL ) ;
	tmpString = static_cast< char* >( headerEntry ) ;
	if( NULL != tmpString ){
		version_ = tmpString ;
	}


	headerGetEntry( hdr , RPMTAG_RELEASE , NULL , &headerEntry , NULL ) ;
	tmpString = static_cast< char* >( headerEntry ) ;
	if( NULL != tmpString ){
		release_ = tmpString ;
	}

	std::ostringstream verTmp ;
	verTmp << version_ << "-" << release_ << std::flush ;

	fullVersion_ = verTmp.str() ;


	headerGetEntry( hdr , RPMTAG_ARCH , NULL , &headerEntry , NULL ) ;
	tmpString = static_cast< char* >( headerEntry ) ;
	if( NULL != tmpString ){
		arch_ = tmpString ;
	}

	headerGetEntry( hdr , RPMTAG_ARCH , NULL , &headerEntry , NULL ) ;
	tmpString = static_cast< char* >( headerEntry ) ;
	if( NULL != tmpString ){
		arch_ = tmpString ;
	}

	if( headerIsEntry( hdr , RPMTAG_SOURCEPACKAGE ) ){
		isSourcePackage_ = true ;
	}

	rpmtsFree( ts ) ;
	fdClose( fileDescriptor ) ;

	/*
	// - - - - - - - - - - - - - - - - - - - -
	std::cout << getFile() << " " << std::flush ;
	std::cout << "\t" << getName() << " " << std::flush ;
	std::cout << "\t" << getArch() << " " << std::flush ;
	std::cout << "\t" << getFullVersion() << " " << std::flush ;
	std::cout << std::endl ;
	// - - - - - - - - - - - - - - - - - - - -
	*/

	return ;

} // RPM::ctor

const std::string& RPM::getFile() const
	throw()
{
	return( file_ ) ;
} // RPM::getFile()


const std::string& RPM::getName() const
	throw()
{
	return( name_ ) ;
} // RPM::getName()


const std::string& RPM::getArch() const
	throw()
{
	return( arch_ ) ;
} // RPM::getArch()


const std::string& RPM::getVersion() const
	throw()
{
	return( version_ ) ;
} // RPM::getVersion()

const std::string& RPM::getRelease() const
	throw()
{
	return( release_ ) ;
} // RPM::getRelease()

const std::string& RPM::getFullVersion() const
	throw()
{
	return( fullVersion_ ) ;
} // RPM::getFullVersion()

const bool RPM::isSourcePackage() const
	throw()
{
	return( isSourcePackage_ ) ;
} // RPM::isSourcePackage()


// - - - - - - - - - - - - - - - - - - - -


bool RPMLess::operator()( const RPM* rpmLeft , const RPM* rpmRight ) const
	throw()
{

		/*
		from the API docs:
 		int rpmvercmp( const char* a, const char* b )  	
		Segmented string compare for version and/or release.

		Returns:
			+1 if a is "newer", 0 if equal, -1 if b is "newer" 

		*/

		/*
		QQQ: double-check logic, based on STL rules for comparators
		*/

	bool result ;

	switch( rpmvercmp( rpmLeft->getVersion().c_str() , rpmRight->getVersion().c_str() ) ){

		case -1: // "rpmRight" version is newer
			result = true ;
			break ;

		case 0: // "rpmLeft" and "rpmRight" are same version, but maybe different release

			if( -1 == rpmvercmp( rpmLeft->getRelease().c_str() , rpmRight->getRelease().c_str() ) ){
				result = true ;
			}else{
				/*
				(sic) STL comparator rules require that a "less than" operator
				return false if the two objects are "equal"
				*/
				result = false ;
			}

			break ;

		case 1: // "rpmLeft" major version is newer
			result = false ;
			break ;
	

	} // switch( rpmvercmp( rpmLeft->getVersion().c_str() , rpmRight->getVersion().c_str() ) )

	return( result ) ;

} // RPMLess::operator()

// - - - - - - - - - - - - - - - - - - - -

Product::Product( const std::string& inId )
	:
	id_( inId ) ,
	list_() ,
	latest_( NULL )
{
	return ;
} // Product::ctor


void Product::push( const RPM* obj ){

	// ... sorting is now handled by set<>
	list_.insert( obj ) ;

	// set<> sorting puts newest RPM in front
	latest_ = *( list_.rbegin() ) ;

	return ;

} // Product::push


Product::~Product()
	throw()
{

	std::for_each(
		list_.begin() ,
		list_.end() ,
		CommonUtil::CleanPointerObjects< RPM >()
	) ;

	return ;

} // Product dtor

const std::string& Product::getId() const
	throw()
{
	return( id_ ) ;
} // Product::getFile()


bool Product::empty() const
	throw()
{
	return( list_.empty() ) ;
} // Product::empty()

int Product::size() const
	throw()
{
	return( list_.size() ) ;
} // Product::size()

const RPM* Product::latestVersion() const
	throw()
{

	// std::cout << "size = " << size() << std::endl ;

	if( NULL == latest_ ){
		// QQQ - throw exception
		std::cerr << "no entry for " << getId() << std::endl ;
		std::exit( 1 ) ;
	}

	return( latest_ ) ;

} // Product::latestVersion()

// - - - - - - - - - - - - - - - - - - - -

extern "C" {
	#include<unistd.h>
	#include<errno.h>
	#include<sys/types.h>	// defs: mode_t
	#include<sys/stat.h>	// stat() , mkdir()

}

#include<iostream>

extern int errno ;

bool Action::createDir( const char* path )
	throw()
{

	bool result = true ;
	struct stat st ;

	if( 0 == stat( path , &st ) ){

		if( ! S_ISDIR( st.st_mode ) ){
			std::cerr << "path \"" << path << "\" exists, but is not a directory" << std::endl ;
			result = false ;
		}

	} else {

		if( 0 != mkdir( path , 0777 ) ){
			std::cerr << "unable to create directory \"" << path << "\": " << strerror( errno ) << std::endl ;
			result = false ;
		}

	}

	return( result ) ;
	
} // Action::createDir()


/*
messy, but preferable to calling C-level strdup()
*/

std::string Action::basename( const std::string& in ){

	// reverse-find "/" and return a new string based on
	// substr( found , end )

	// if "/" not found, return the original string

	std::string::size_type found = in.rfind( "/" ) ;

	if( std::string::npos == found ){

		return( in ) ;

	} else {

		std::string result( in.substr( found + 1 , std::string::npos ) ) ;
		return( result ) ;

	}

} // Action::basename()

void ListAction::operator()( const Product& product )
	throw()
{

	std::cout << product.latestVersion()->getFile() << std::endl ;

	return ;

} // ListAction::operator()()


void PrintAction::operator()( const Product& product )
	throw()
{

	const RPM* rpm = product.latestVersion() ;

	// QQQ - add operator<<() for RPM
	std::cout << rpm->getName() << "\t" << rpm->getFile() << std::endl ;

	return ;

} // PrintAction::operator()()


HardlinkAction::HardlinkAction( const std::string& path )
	:
	path_( path )
{
	return ;
} // HardlinkAction::ctor

void HardlinkAction::operator()( const Product& product )
	throw()
{

	const RPM* rpm = product.latestVersion() ;

	if( ! Action::createDir( path_.c_str() ) ){
		std::cerr << "not copying \"" << rpm->getFile() << "\"" << std::endl ;
		return ;
	}

	std::string justFile = basename( rpm->getFile().c_str() ) ;

	buf_.str( "" ) ;
	buf_ << path_ << "/" << justFile << std::flush ;

	if( 0 != link( rpm->getFile().c_str() , buf_.str().c_str() ) ){
		std::cerr << "unable to link \"" << rpm->getFile() << "\" => \"" << buf_.str() << ": " << strerror( errno ) << std::endl ;
	}

	return ;

} // HardlinkAction::operator()



SoftlinkAction::SoftlinkAction( const std::string& path )
	:
	path_( path )
{
} // SoftlinkAction::ctor

void SoftlinkAction::operator()( const Product& product )
	throw()
{

	const RPM* rpm = product.latestVersion() ;

	if( ! Action::createDir( path_.c_str() ) ){
		std::cerr << "not copying \"" << rpm->getFile() << "\"" << std::endl ;
		return ;
	}

	std::string justFile = basename( rpm->getFile().c_str() ) ;

	buf_.str( "" ) ;
	buf_ << path_ << "/" << justFile << std::flush ;

	if( 0 != symlink( rpm->getFile().c_str() , buf_.str().c_str() ) ){
		std::cerr << "unable to link \"" << rpm->getFile() << "\" => \"" << buf_.str() << ": " << strerror( errno ) << std::endl ;
	}

	return ;

} // SoftlinkAction::operator()



// - - - - - - - - - - - - - - - - - - - -

#include<map>
#include<string>
#include<sstream>
#include<vector>



ProductListIterator::ProductListIterator( ProductListInternalMap::iterator stop )
	:
	current_( stop ) ,
	stop_( stop )
{

	/*
	not a typo -- current_ is used in operator==() so it should be the same for
	both current_ and stop_, since this overload of the ctor creates the "end"
	iterator
	*/

	return ;

} // ProductListIterator::ctor for "end" value

ProductListIterator::ProductListIterator( ProductListInternalMap::iterator start , ProductListInternalMap::iterator stop )
	:
	current_( start ) ,
	stop_( stop )
{
	return ;
} // ProductListIterator::ctor for "begin" value

ProductListIterator::self_type& ProductListIterator::operator++(){

	if( current_ == stop_ ){
		// QQQ - ?throw exception? or let operator shoot self in foot?
	}

	++current_ ;

	return( *this ) ;

} // ProductListIterator::operator++ (pre)

ProductListIterator::self_type ProductListIterator::operator++(int){

	self_type result( *this ) ;
	++( *this ) ;

	return( result ) ;

} // ProductListIterator::operator++ (post)


bool ProductListIterator::operator==( const self_reference other ){
	return(
		other.current_ == this->current_
	) ;
} // ProductListIterator::oeprator==

bool ProductListIterator::operator!=( const self_reference other ){
	return( ! ( (*this) == other ) ) ;
} // ProductListIterator::operator!=


ProductListIterator::reference ProductListIterator::operator*(){
	return( current_->second ) ;
} // ProductListIterator::operator*()


ProductListIterator::reference ProductListIterator::operator->(){
	return( *( *this ) ) ;
} // ProductListIterator::operator->



ProductList::ProductList()
	:
	map_()
{
	return ;
} // ProductList::ctor


void ProductList::add( const std::string path ){

	/*
	QQQ: take a "fail-on-any-error" flag from commandline?
	The exception doesn't leave this function...
	
	Perhaps this exception should be propogated up the stack
	such that something closer to main() (ergo, something closer
	to AppConfig) can find it.
	*/

	RPM* result = NULL ;

	try{

		result = new RPM( path ) ;
		add( result ) ;

	}catch( AppException& e ){
		std::cerr << "problem loading \"" << path << "\": " << e.what() << std::endl ;
	}


	return ;

} // add()


void ProductList::add( const RPM* rpm ) {

	// QQQ: wrap this logic in a filter object that protects
	// the insertion
	if( rpm->isSourcePackage() ){
		return ;
	}

	std::ostringstream keyBuf ;
	keyBuf << rpm->getName() << ":" << rpm->getArch() ;

	std::string key = keyBuf.str() ;

	ProductListInternalMap::iterator found = map_.find( key ) ;

	/*
	strange insert is due to double-free of RPM* objects in Product dtor...
	must find another way around this.  (Make Product objects pointers, maybe?)
	*/

	if( map_.end() == found ){
		Product pr( key ) ;
		// pr.push( rpm ) ;
		map_.insert( ProductList::value_type( pr.getId() , pr ) ) ;

		map_.find( key )->second.push( rpm ) ;

	} else {
		found->second.push( rpm ) ;
	}

	return ;

} // ProductList::add( const RPM* )

ProductList::iterator ProductList::begin() {

	ProductListIterator result( map_.begin() , map_.end() ) ;

	return( result ) ;
	
} // ProductList::begin()

ProductList::iterator ProductList::end()  {

	ProductListIterator result( map_.end() ) ;

	return( result ) ;

} // ProductList::end()

int ProductList::size()  {

	return( map_.size() ) ;

} // ProductList::size()

// - - - - - - - - - - - - - - - - - - - -
// - - - - - - - - - - - - - - - - - - - -
